package wacc

import scala.collection.mutable.LinkedHashMap
import scala.collection.mutable.ListBuffer
import cats.data.Chain
import cats.syntax.foldable._

object asmGenerator {
  import microWacc._
  import assemblyIR._
  import wacc.types._
  import lexer.escapedChars

  val RAX = Register(RegSize.R64, RegName.AX)
  val EAX = Register(RegSize.E32, RegName.AX)
  val ESP = Register(RegSize.E32, RegName.SP)
  val EDX = Register(RegSize.E32, RegName.DX)
  val RDI = Register(RegSize.R64, RegName.DI)
  val RIP = Register(RegSize.R64, RegName.IP)
  val RBP = Register(RegSize.R64, RegName.BP)
  val RSI = Register(RegSize.R64, RegName.SI)
  val RDX = Register(RegSize.R64, RegName.DX)
  val RCX = Register(RegSize.R64, RegName.CX)
  val R8 = Register(RegSize.R64, RegName.Reg8)
  val R9 = Register(RegSize.R64, RegName.Reg9)

  val _8_BIT_MASK = 0xff

  def generateAsm(microProg: Program): List[AsmLine] = {
    given stack: Stack = Stack()
    given strings: ListBuffer[String] = ListBuffer[String]()
    given labelGenerator: LabelGenerator = LabelGenerator()
    val Program(funcs, main) = microProg

    val progAsm =
      Chain.one(LabelDef("main")) ++
        funcPrologue() ++
        Chain(stack.align()) ++
        main.foldLeft(Chain.empty[AsmLine])(_ ++ generateStmt(_)) ++
        Chain.one(Move(RAX, ImmediateVal(0))) ++
        funcEpilogue() ++
        generateBuiltInFuncs()

    val strDirs = strings.toList.zipWithIndex.foldLeft(Chain.empty[AsmLine]) {
      case (acc, (str, i)) =>
        acc ++ Chain(
          Directive.Int(str.size),
          LabelDef(s".L.str$i"),
          Directive.Asciz(str.escaped)
        )
    }

    val finalChain = Chain(
      Directive.IntelSyntax,
      Directive.Global("main"),
      Directive.RoData
    ) ++ strDirs ++ Chain.one(Directive.Text) ++ progAsm

    finalChain.toList

  }

  def wrapFunc(labelName: String, funcBody: Chain[AsmLine])(using
      stack: Stack,
      strings: ListBuffer[String]
  ): Chain[AsmLine] = {
    Chain.one(LabelDef(labelName)) ++
      funcPrologue() ++
      funcBody ++
      funcEpilogue()
  }

  def generateBuiltInFuncs()(using
      stack: Stack,
      strings: ListBuffer[String]
  ): Chain[AsmLine] = {
    wrapFunc(
      labelGenerator.getLabel(Builtin.Exit),
      Chain(stack.align(), assemblyIR.Call(CLibFunc.Exit))
    ) ++
      wrapFunc(
        labelGenerator.getLabel(Builtin.Printf),
        Chain(
          stack.align(),
          assemblyIR.Call(CLibFunc.PrintF),
          Move(RDI, ImmediateVal(0)),
          assemblyIR.Call(CLibFunc.Fflush)
        )
      ) ++
      wrapFunc(
        labelGenerator.getLabel(Builtin.Malloc),
        Chain.one(
          stack.align()
        )
      ) ++
      wrapFunc(labelGenerator.getLabel(Builtin.Free), Chain.empty) ++
      wrapFunc(
        labelGenerator.getLabel(Builtin.Read),
        Chain(
          stack.align(),
          stack.reserve(),
          stack.push(RSI),
          Load(RSI, stack.head),
          assemblyIR.Call(CLibFunc.Scanf),
          stack.pop(RAX),
          stack.drop()
        )
      )
  }

  def generateStmt(
      stmt: Stmt
  )(using stack: Stack, strings: ListBuffer[String]): Chain[AsmLine] =
    stmt match {
      case Assign(lhs, rhs) =>
        var dest: () => IndexAddress =
          () => IndexAddress(RAX, 0) // gets overrwitten
        (lhs match {
          case ident: Ident =>
            dest = stack.accessVar(ident)
            if (!stack.contains(ident)) {
              Chain.one(stack.reserve(ident))
            } else Chain.empty
          // TODO lhs = arrayElem
          case _ =>
            // dest = ???
            Chain.empty
        }) ++
          evalExprOntoStack(rhs) ++
          Chain(
            stack.pop(RAX),
            Move(dest(), RAX)
          )
      case If(cond, thenBranch, elseBranch) => {
        val elseLabel = labelGenerator.getLabel()
        val endLabel = labelGenerator.getLabel()
        evalExprOntoStack(cond) ++
          Chain.fromSeq(
            List(
              Compare(stack.head(SizeDir.Word), ImmediateVal(0)),
              stack.drop(),
              Jump(LabelArg(elseLabel), Cond.Equal)
            )
          ) ++
          Chain.fromSeq(thenBranch).flatMap(generateStmt) ++
          Chain.fromSeq(List(Jump(LabelArg(endLabel)), LabelDef(elseLabel))) ++
          Chain.fromSeq(elseBranch).flatMap(generateStmt) ++
          Chain.one(LabelDef(endLabel))
      }
      case While(cond, body) => {
        val startLabel = labelGenerator.getLabel()
        val endLabel = labelGenerator.getLabel()
        Chain.one(LabelDef(startLabel)) ++
          evalExprOntoStack(cond) ++
          Chain(
            Compare(stack.head(SizeDir.Word), ImmediateVal(0)),
            stack.drop(),
            Jump(LabelArg(endLabel), Cond.Equal)
          ) ++
          Chain.fromSeq(body).flatMap(generateStmt) ++
          Chain(Jump(LabelArg(startLabel)), LabelDef(endLabel))
      }
      case microWacc.Return(expr) =>
        evalExprOntoStack(expr) ++
          Chain(stack.pop(RAX), assemblyIR.Return())
      case call: microWacc.Call => generateCall(call)
    }

  def evalExprOntoStack(expr: Expr)(using
      stack: Stack,
      strings: ListBuffer[String]
  ): Chain[AsmLine] = {
    val out = expr match {
      case IntLiter(v) =>
        Chain.one(stack.push(ImmediateVal(v)))
      case CharLiter(v) =>
        Chain.one(stack.push(ImmediateVal(v.toInt)))
      case ident: Ident =>
        Chain.one(stack.push(stack.accessVar(ident)()))
      case ArrayLiter(elems) =>
        expr.ty match {
          case KnownType.String =>
            strings += elems.map {
              case CharLiter(v) => v
              case _            => ""
            }.mkString
            Chain(
              Load(
                RAX,
                IndexAddress(
                  RIP,
                  LabelArg(s".L.str${strings.size - 1}")
                )
              ),
              stack.push(RAX)
            )
          // TODO other array types
          case _ => Chain.empty
        }
      case BoolLiter(v)              => Chain.one(stack.push(ImmediateVal(if (v) 1 else 0)))
      case NullLiter()               => Chain.one(stack.push(ImmediateVal(0)))
      case ArrayElem(value, indices) => Chain.empty
      case UnaryOp(x, op) =>
        evalExprOntoStack(x) ++
          (op match {
            // TODO: chr and ord are TYPE CASTS. They do not change the internal value,
            //       but will need bound checking e.t.c.
            case UnaryOperator.Chr | UnaryOperator.Ord | UnaryOperator.Len => Chain.empty
            case UnaryOperator.Negate =>
              Chain.one(
                Negate(stack.head(SizeDir.Word))
              )
            case UnaryOperator.Not =>
              evalExprOntoStack(x) ++
                Chain.one(
                  Xor(stack.head(SizeDir.Word), ImmediateVal(1))
                )

          })
      case BinaryOp(x, y, op) =>
        op match {
          case BinaryOperator.Add =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              Chain(
                stack.pop(RAX),
                Add(stack.head(SizeDir.Word), EAX)
                // TODO OVERFLOWING
              )
          case BinaryOperator.Sub =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              Chain(
                stack.pop(RAX),
                Subtract(stack.head(SizeDir.Word), EAX)
                // TODO OVERFLOWING
              )
          case BinaryOperator.Mul =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              Chain(
                stack.pop(RAX),
                Multiply(EAX, stack.head(SizeDir.Word)),
                stack.drop(),
                stack.push(RAX)
                // TODO OVERFLOWING
              )
          case BinaryOperator.Div =>
            evalExprOntoStack(y) ++
              evalExprOntoStack(x) ++
              Chain(
                stack.pop(RAX),
                Divide(stack.head(SizeDir.Word)),
                stack.drop(),
                stack.push(RAX)
                // TODO CHECK DIVISOR IS NOT 0
              )
          case BinaryOperator.Mod =>
            evalExprOntoStack(y) ++
              evalExprOntoStack(x) ++
              Chain(
                stack.pop(RAX),
                Divide(stack.head(SizeDir.Word)),
                stack.drop(),
                stack.push(RDX)
                // TODO CHECK DIVISOR IS NOT 0
              )
          case BinaryOperator.Eq =>
            generateComparison(x, y, Cond.Equal)
          case BinaryOperator.Neq =>
            generateComparison(x, y, Cond.NotEqual)
          case BinaryOperator.Greater =>
            generateComparison(x, y, Cond.Greater)
          case BinaryOperator.GreaterEq =>
            generateComparison(x, y, Cond.GreaterEqual)
          case BinaryOperator.Less =>
            generateComparison(x, y, Cond.Less)
          case BinaryOperator.LessEq =>
            generateComparison(x, y, Cond.LessEqual)
          case BinaryOperator.And =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              Chain(
                stack.pop(RAX),
                And(stack.head(SizeDir.Word), EAX)
              )
          case BinaryOperator.Or =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              Chain(
                stack.pop(RAX),
                Or(stack.head(SizeDir.Word), EAX)
              )
        }
      case call: microWacc.Call =>
        generateCall(call) ++
          Chain.one(stack.push(RAX))
    }
    if out.isEmpty then Chain.one(stack.push(ImmediateVal(0))) else out
  }

  def generateCall(call: microWacc.Call)(using
      stack: Stack,
      strings: ListBuffer[String]
  ): Chain[AsmLine] = {
    val argRegs = List(RDI, RSI, RDX, RCX, R8, R9)
    val microWacc.Call(target, args) = call

    val regMoves = argRegs
      .zip(args)
      .map { (reg, expr) =>
        evalExprOntoStack(expr) ++
          Chain.one(stack.pop(reg))
      }
      .combineAll

    val stackPushes = args.drop(argRegs.size).map(evalExprOntoStack).combineAll

    regMoves ++
      stackPushes ++
      Chain.one(assemblyIR.Call(LabelArg(labelGenerator.getLabel(target)))) ++
      (if (args.size > argRegs.size) Chain.one(stack.drop(args.size - argRegs.size))
       else Chain.empty)
  }

  def generateComparison(x: Expr, y: Expr, cond: Cond)(using
      stack: Stack,
      strings: ListBuffer[String]
  ): Chain[AsmLine] = {
    evalExprOntoStack(x) ++
      evalExprOntoStack(y) ++
      Chain(
        stack.pop(RAX),
        Compare(stack.head(SizeDir.Word), EAX),
        Set(Register(RegSize.Byte, RegName.AL), cond),
        And(RAX, ImmediateVal(_8_BIT_MASK)),
        stack.drop(),
        stack.push(RAX)
      )
  }

  // Missing a sub instruction but dont think we need it
  def funcPrologue()(using stack: Stack): Chain[AsmLine] = {
    Chain(
      stack.push(RBP),
      Move(RBP, Register(RegSize.R64, RegName.SP))
    )
  }

  def funcEpilogue()(using stack: Stack): Chain[AsmLine] = {
    Chain(
      Move(Register(RegSize.R64, RegName.SP), RBP),
      stack.pop(RBP),
      assemblyIR.Return()
    )
  }

  class LabelGenerator {
    var labelVal = -1
    def getLabel(): String = {
      labelVal += 1
      s".L$labelVal"
    }
    def getLabel(target: CallTarget): String = target match {
      case Ident(v, _)   => s"wacc_$v"
      case Builtin(name) => s"_$name"
    }
  }

  class Stack {
    private val stack = LinkedHashMap[Expr | Int, Int]()
    private val RSP = Register(RegSize.R64, RegName.SP)

    def next: Int = stack.size + 1
    def push(expr: Expr, src: Src): AsmLine = {
      stack += expr -> next
      Push(src)
    }
    def push(src: Src): AsmLine = {
      stack += stack.size -> next
      Push(src)
    }
    def pop(dest: Src): AsmLine = {
      stack.remove(stack.last._1)
      Pop(dest)
    }
    def reserve(ident: Ident): AsmLine = {
      stack += ident -> next
      Subtract(RSP, ImmediateVal(8))
    }
    def reserve(n: Int = 1): AsmLine = {
      (1 to n).foreach(_ => stack += stack.size -> next)
      Subtract(RSP, ImmediateVal(n * 8))
    }
    def drop(n: Int = 1): AsmLine = {
      (1 to n).foreach(_ => stack.remove(stack.last._1))
      Add(RSP, ImmediateVal(n * 8))
    }
    def accessVar(ident: Ident): () => IndexAddress = () => {
      IndexAddress(RSP, (stack.size - stack(ident)) * 8)
    }
    def head: MemLocation = MemLocation(RSP)
    def head(size: SizeDir): MemLocation = MemLocation(RSP, size)
    def contains(ident: Ident): Boolean = stack.contains(ident)
    // TODO: Might want to actually properly handle this with the LinkedHashMap too
    def align(): AsmLine = And(RSP, ImmediateVal(-16))
  }

  private val escapedCharsMapping = escapedChars.map { case (k, v) => v -> s"\\$k" }
  extension (s: String) {
    private def escaped: String =
      s.flatMap(c => escapedCharsMapping.getOrElse(c, c.toString))
  }
}
